# --------------------------
# File: mcp_pinot/server.py
# --------------------------
import asyncio
import logging
from typing import Any

import mcp.types as types
from mcp.server import NotificationOptions, Server
from mcp.server.models import InitializationOptions
import mcp.server.stdio
import sys
from mcp_pinot.utils.pinot_client import (
    PINOT_CONTROLLER_URL,
    PINOT_USERNAME,
    PINOT_PASSWORD,
    PINOT_TOKEN,
    PINOT_USE_MSQE,
    HEADERS,
    conn,
    Pinot
)

from mcp_pinot.prompts import PROMPT_TEMPLATE

logger = logging.getLogger("pinot_mcp_claude")
logger.setLevel(logging.INFO)

# Use the imported Pinot class and connection values
pinot_instance = Pinot()

async def main():
    logger.info("Starting Pinot MCP Server")
    server = Server("pinot_mcp_claude")

    @server.list_prompts()
    async def handle_list_prompts() -> list[types.Prompt]:
        logger.debug("Handling list_prompts request")
        return [
            types.Prompt(
                name="pinot-query",
                description="A prompt to Query the pinot database with an Pinot MCP Server + Claude",
                arguments=[],
            )
        ]

    @server.get_prompt()
    async def handle_get_prompt(name: str, arguments: dict[str, str] | None) -> types.GetPromptResult:
        if name != "pinot-query":
            raise ValueError(f"Unknown prompt: {name}")
        return types.GetPromptResult(
            description="Pinot query assistance template",
            messages=[
                types.PromptMessage(
                    role="user",
                    content=types.TextContent(type="text", text=PROMPT_TEMPLATE.strip()),
                )
            ],
        )

    @server.list_tools()
    async def handle_list_tools() -> list[types.Tool]:
        return [
            types.Tool(
                name="read-query",
                description="Execute a SELECT query on the Pinot database",
                inputSchema={
                    "type": "object",
                    "properties": {
                        "query": {"type": "string", "description": "SELECT SQL query to execute"},
                    },
                    "required": ["query"],
                },
            ),
            types.Tool(
                name="list-tables",
                description="List all tables in Pinot",
                inputSchema={"type": "object", "properties": {}},
            ),
            types.Tool(
                name="table-details",
                description="Get table size details",
                inputSchema={
                    "type": "object",
                    "properties": {
                        "tableName": {"type": "string", "description": "Table name"},
                    },
                    "required": ["tableName"],
                },
            ),
            types.Tool(
                name="segment-list",
                description="List segments for a table",
                inputSchema={
                    "type": "object",
                    "properties": {
                        "tableName": {"type": "string", "description": "Table name"},
                    },
                    "required": ["tableName"],
                },
            ),
            types.Tool(
                name="index-column-details",
                description="Get index/column details for a segment",
                inputSchema={
                    "type": "object",
                    "properties": {
                        "tableName": {"type": "string"},
                        "segmentName": {"type": "string"},
                    },
                    "required": ["tableName", "segmentName"],
                },
            ),
            types.Tool(
                name="segment-metadata-details",
                description="Get metadata for segments of a table",
                inputSchema={
                    "type": "object",
                    "properties": {
                        "tableName": {"type": "string"},
                    },
                    "required": ["tableName"],
                },
            ),
            types.Tool(
                name="tableconfig-schema-details",
                description="Get table config and schema",
                inputSchema={
                    "type": "object",
                    "properties": {
                        "tableName": {"type": "string"},
                    },
                    "required": ["tableName"],
                },
            ),
        ]

    @server.call_tool()
    async def handle_call_tool(
        name: str, arguments: dict[str, Any] | None
    ) -> list[types.TextContent | types.ImageContent | types.EmbeddedResource]:
        """Handle tool execution requests"""
        try:
            if name == "read-query":
                if not arguments["query"].strip().upper().startswith("SELECT"):
                    raise ValueError("Only SELECT queries are allowed for read-query")
                results = pinot_instance._execute_query(query=arguments["query"])
                return [types.TextContent(type="text", text=str(results))]

            elif name == "table-details":
                results = pinot_instance._get_table_detail(tableName=arguments["tableName"])
                return [types.TextContent(type="text", text=str(results))]

            elif name == "segment-list":
                results = pinot_instance._get_segments(tableName=arguments["tableName"])
                return [types.TextContent(type="text", text=str(results))]

            elif name == "index-column-details":
                results = pinot_instance._get_index_column_detail(
                    tableName=arguments["tableName"],
                    segmentName=arguments["segmentName"]
                )
                return [types.TextContent(type="text", text=str(results))]

            elif name == "segment-metadata-details":
                results = pinot_instance._get_segment__metadata_detail(tableName=arguments["tableName"])
                return [types.TextContent(type="text", text=str(results))]

            elif name == "tableconfig-schema-details":
                results = pinot_instance._get_tableconfig_schema_detail(tableName=arguments["tableName"])
                return [types.TextContent(type="text", text=str(results))]

            elif name == "list-tables":
                results = pinot_instance._get_tables()
                return [types.TextContent(type="text", text=str(results))]

            else:
                raise ValueError(f"Unknown tool: {name}")

        except Exception as e:
            return [types.TextContent(type="text", text=f"Error: {str(e)}")]

    try:
        async with mcp.server.stdio.stdio_server() as (read_stream, write_stream):
            logger.info("Server running with stdio transport")
            await server.run(
                read_stream,
                write_stream,
                InitializationOptions(
                    server_name="pinot_mcp_claude",
                    server_version="0.1.0",
                    capabilities=server.get_capabilities(
                        notification_options=NotificationOptions(),
                        experimental_capabilities={},
                    ),
                ),
            )
    except Exception as e:
        import traceback
        logger.error(f"Error running MCP server: {e}")
        logger.error(traceback.format_exc())
        print(f"Error running MCP server: {e}", file=sys.stderr)
        print(traceback.format_exc(), file=sys.stderr)
        raise

if __name__ == "__main__":
    import asyncio
    asyncio.run(main())



import pytest
from unittest.mock import patch, AsyncMock, MagicMock
import mcp.types as types
from mcp.server import NotificationOptions
from mcp_pinot.server import main, pinot_instance

@pytest.fixture
def mock_server():
    """Fixture to mock the Server class."""
    with patch("mcp_pinot.server.Server") as mock_server_class:
        mock_server = MagicMock()
        # Set up async mock methods
        mock_server.list_prompts = AsyncMock()
        mock_server.get_prompt = AsyncMock()
        mock_server.list_tools = AsyncMock()
        mock_server.call_tool = AsyncMock()
        mock_server_class.return_value = mock_server
        yield mock_server

@pytest.mark.asyncio
async def test_handle_list_prompts(mock_server):
    """Test the handle_list_prompts function."""
    # Set up the mock return value
    mock_server.list_prompts.return_value = [
        types.Prompt(
            name="pinot-query",
            description="Query the pinot database",
            arguments=[]
        )
    ]
    
    # Call the function
    result = await mock_server.list_prompts()
    
    # Check the result
    assert isinstance(result, list)
    assert len(result) == 1
    assert result[0].name == "pinot-query"
    assert "Query the pinot database" in result[0].description
    assert isinstance(result[0].arguments, list)

@pytest.mark.asyncio
async def test_handle_get_prompt_valid(mock_server):
    """Test the handle_get_prompt function with a valid prompt name."""
    # Set up the mock return value
    mock_server.get_prompt.return_value = types.GetPromptResult(
        description="Pinot query assistance template",
        messages=[
            types.PromptMessage(
                role="user",
                content=types.TextContent(type="text", text="pinot query template"),
            )
        ]
    )
    
    # Call the function
    result = await mock_server.get_prompt("pinot-query", None)
    
    # Check the result
    assert isinstance(result, types.GetPromptResult)
    assert "pinot" in result.messages[0].content.text.lower()

@pytest.mark.asyncio
async def test_handle_get_prompt_invalid(mock_server):
    """Test the handle_get_prompt function with an invalid prompt name."""
    # Set up the mock to raise an exception
    mock_server.get_prompt.side_effect = ValueError("Unknown prompt")
    
    # Call the function with an invalid prompt name
    with pytest.raises(ValueError, match="Unknown prompt"):
        await mock_server.get_prompt("invalid-prompt", None)

@pytest.mark.asyncio
async def test_handle_list_tools(mock_server):
    """Test the handle_list_tools function."""
    # Set up the mock return value
    mock_server.list_tools.return_value = [
        types.Tool(
            name="test_tool",
            description="Test tool description",
            inputSchema={
                "type": "object",
                "properties": {},
            }
        )
    ]
    
    # Call the function
    result = await mock_server.list_tools()
    
    # Check the result
    assert isinstance(result, list)
    assert len(result) > 0
    
    # Check that each tool has the required attributes
    for tool in result:
        assert hasattr(tool, "name")
        assert hasattr(tool, "description")
        assert hasattr(tool, "inputSchema")

@pytest.mark.asyncio
async def test_handle_call_tool(mock_server):
    """Test the handle_call_tool function."""
    # Set up the mock return value with the required 'type' field
    mock_server.call_tool.return_value = [
        types.TextContent(type="text", text="Test result")
    ]
    
    # Call the function
    result = await mock_server.call_tool("run_select_query", {"sql": "SELECT * FROM my_table"})
    
    # Check the result
    assert isinstance(result, list)
    assert len(result) == 1
    assert isinstance(result[0], types.TextContent)
    assert result[0].text == "Test result"

@pytest.mark.asyncio
async def test_handle_call_tool_invalid_tool(mock_server):
    """Test the handle_call_tool function with an invalid tool name."""
    # Set up the mock to raise an exception
    mock_server.call_tool.side_effect = ValueError("Unknown tool")
    
    # Call the function with an invalid tool name
    with pytest.raises(ValueError, match="Unknown tool"):
        await mock_server.call_tool("invalid_tool", {})

@pytest.mark.asyncio
async def test_main_function():
    """Test the main function."""
    # Mock the Server class and its methods
    with patch("mcp_pinot.server.Server") as mock_server_class:
        mock_server = MagicMock()
        mock_server.run = AsyncMock()
        
        # Set up the get_capabilities mock to return a valid ServerCapabilities object
        mock_server.get_capabilities.return_value = types.ServerCapabilities(
            supportsPrompts=True,
            supportsTools=True,
            supportsNotifications=True,
            supportsExperimentalCapabilities=True,
        )
        
        mock_server_class.return_value = mock_server
        
        # Mock the stdio_server context manager
        with patch("mcp.server.stdio.stdio_server") as mock_stdio_server:
            mock_read_stream = AsyncMock()
            mock_write_stream = AsyncMock()
            mock_stdio_server.return_value.__aenter__.return_value = (mock_read_stream, mock_write_stream)
            
            # Call the main function
            await main()
            
            # Check that the server was run with the correct arguments
            mock_server.run.assert_called_once() 



PROMPT_TEMPLATE_V1 = """
Apache Pinot is a real-time distributed OLAP datastore purpose-built for low-latency, high-throughput analytics, and perfect for user-facing analytical workloads.

Apache Pinot™ is a real-time distributed online analytical processing (OLAP) datastore. Use Pinot to ingest and immediately query data from streaming or batch data sources (including, Apache Kafka, Amazon Kinesis, Hadoop HDFS, Amazon S3, Azure ADLS, and Google Cloud Storage)
. You can get a more detailed description and documentation about Apache pinot using the dosc ar "https://docs.pinot.apache.org/" tool.

The assistants goal is to get insights from a Pinot Workspace. To get those insights we will leverage this server to interact with pinot deployment. The user is a business decision maker with no previous knowledge of the data structure or insights inside the pinot Workspace.

You job is to simply execute READ Only Select queries from pinot using the python driver and help user visualise the data
"""

PROMPT_TEMPLATE_V2 = """
You are an AI analyst assistant for Apache Pinot, a real-time distributed OLAP datastore. Your role is to help users analyze Pinot data using natural language queries, convert these queries to SQL, suggest data visualizations, and ask clarifying questions when needed.


You have access to the following tools to assist in your analysis:

1. read-query: Execute a SQL query on Pinot and return the results
2. list-tables: List all available tables in Pinot
3. list-schema: List the schema for a specific table
4. table-details: Get detailed information about a specific table
5. index-column-details: Get index details for a specific column in a table
6. segment-list: List all segments for a specific table
7. segment-metadata-details: Get metadata details for a specific segment
8. tableconfig-schema-details: Get combined table configuration and schema details

When a user provides a query, follow these steps:

1. Analyze the user's natural language query and identify the key elements (e.g., table, columns, filters, time range).

2. Based on the Pinot schema and the user's query, determine which table(s) and columns are relevant to the analysis.

3. Convert the natural language query into a SQL query that can be executed on Pinot. Ensure that the SQL query is optimized for Pinot's capabilities and follows best practices.

4. If the query is ambiguous or lacks necessary information, formulate clarifying questions to ask the user. Present these questions clearly and concisely.

5. Suggest appropriate data visualizations based on the nature of the query and the expected results. Consider charts, graphs, or other visual representations that would effectively communicate the insights.

6. If additional information about the schema, table configuration, or indexes is needed to optimize the query or provide better recommendations, use the appropriate tools (e.g., list-schema, table-details, index-column-details) to gather this information.

7. Present your findings in the following format:

<analysis>
<sql_query>
[Insert the converted SQL query here]
</sql_query>

<explanation>
[Provide a brief explanation of how the SQL query addresses the user's question]
</explanation>

<clarifying_questions>
[List any clarifying questions, if needed]
</clarifying_questions>

<visualization_suggestions>
[Provide suggestions for data visualization]
</visualization_suggestions>

<additional_insights>
[Include any additional insights or recommendations based on your analysis]
</additional_insights>
</analysis>

Remember to always prioritize clarity and accuracy in your responses. If you're unsure about any aspect of the query or analysis, it's better to ask for clarification than to make assumptions.
"""

PROMPT_TEMPLATE = PROMPT_TEMPLATE_V2

def generate_prompt(topic: str) -> str:
    return PROMPT_TEMPLATE.format(topic=topic)



import pytest
from unittest.mock import patch, MagicMock
import pandas as pd
from mcp_pinot.utils.pinot_client import Pinot, conn

# Mock data for testing
MOCK_TABLE_DATA = [
    {"id": 1, "name": "Table 1"},
    {"id": 2, "name": "Table 2"}
]

MOCK_QUERY_RESULT = pd.DataFrame([
    {"id": 1, "name": "Test 1"},
    {"id": 2, "name": "Test 2"}
])

@pytest.fixture
def mock_connection():
    """Fixture to mock the Pinot connection."""
    with patch("mcp_pinot.utils.pinot_client.connect") as mock_connect:
        mock_cursor = MagicMock()
        mock_cursor.description = [("id",), ("name",)]
        mock_cursor.__iter__.return_value = [(1, "Test 1"), (2, "Test 2")]
        mock_connect.return_value.cursor.return_value = mock_cursor
        yield mock_connect

@pytest.fixture
def mock_requests():
    """Fixture to mock the requests module."""
    with patch("mcp_pinot.utils.pinot_client.requests") as mock_req:
        mock_response = MagicMock()
        mock_response.json.return_value = {"tables": ["table1", "table2"]}
        mock_req.get.return_value = mock_response
        yield mock_req

def test_pinot_init():
    """Test that Pinot class initializes correctly."""
    pinot = Pinot()
    assert isinstance(pinot.insights, list)
    assert len(pinot.insights) == 0

@patch("mcp_pinot.utils.pinot_client.conn")
def test_execute_query(mock_conn, mock_connection):
    """Test the execute_query function."""
    # Set up the mock cursor
    mock_cursor = MagicMock()
    mock_cursor.description = [("id",), ("name",)]
    mock_cursor.__iter__.return_value = [(1, "Test 1"), (2, "Test 2")]
    mock_conn.cursor.return_value = mock_cursor
    
    pinot = Pinot()
    result = pinot._execute_query("SELECT * FROM my_table")
    assert isinstance(result, list)
    assert len(result) == 2
    assert result[0]["id"] == 1
    assert result[0]["name"] == "Test 1"

@patch("mcp_pinot.utils.pinot_client.conn")
def test_execute_query_empty_result(mock_conn, mock_connection):
    """Test execute_query with an empty result set."""
    mock_cursor = MagicMock()
    mock_cursor.description = [("id",), ("name",)]
    mock_cursor.__iter__.return_value = []
    mock_conn.cursor.return_value = mock_cursor
    
    pinot = Pinot()
    result = pinot._execute_query("SELECT * FROM my_table WHERE id = 999")
    assert isinstance(result, list)
    assert len(result) == 0

@patch("mcp_pinot.utils.pinot_client.conn")
def test_execute_query_with_error(mock_conn, mock_connection):
    """Test execute_query with a database error."""
    mock_conn.cursor.side_effect = Exception("Database error")
    
    pinot = Pinot()
    with pytest.raises(Exception, match="Database error"):
        pinot._execute_query("SELECT * FROM my_table")

def test_pinot_get_tables(mock_requests):
    """Test the _get_tables method."""
    pinot = Pinot()
    tables = pinot._get_tables()
    assert isinstance(tables, list)
    assert "table1" in tables
    assert "table2" in tables

def test_pinot_get_table_detail(mock_requests):
    """Test the _get_table_detail method."""
    mock_requests.get.return_value.json.return_value = {
        "tableName": "test_table",
        "columnCount": 5
    }
    
    pinot = Pinot()
    detail = pinot._get_table_detail("test_table")
    assert isinstance(detail, dict)
    assert detail["tableName"] == "test_table"
    assert detail["columnCount"] == 5

def test_pinot_list_tools():
    """Test the list_tools method."""
    pinot = Pinot()
    tools = pinot.list_tools()
    assert isinstance(tools, list)
    assert len(tools) > 0
    
    # Check that each tool has the required attributes
    for tool in tools:
        assert hasattr(tool, "name")
        assert hasattr(tool, "description")
        assert hasattr(tool, "inputSchema") 


import logging
from typing import Any
import pandas as pd
import requests
import mcp.types as types
from pinotdb import connect
import os
from dotenv import load_dotenv

# Load environment variables from .env file
load_dotenv()

logger = logging.getLogger("pinot_mcp_claude")

# Get configuration from environment variables
PINOT_CONTROLLER_URL = os.getenv("PINOT_CONTROLLER_URL")
PINOT_BROKER_HOST = os.getenv("PINOT_BROKER_HOST")
PINOT_BROKER_PORT = int(os.getenv("PINOT_BROKER_PORT", "443"))
PINOT_BROKER_SCHEME = os.getenv("PINOT_BROKER_SCHEME", "https")
PINOT_USERNAME = os.getenv("PINOT_USERNAME")
PINOT_PASSWORD = os.getenv("PINOT_PASSWORD")
PINOT_USE_MSQE = os.getenv("PINOT_USE_MSQE", "false").lower() == "true"
PINOT_TOKEN = os.getenv("PINOT_TOKEN", "")

HEADERS = {
    "accept": "application/json",
}
if PINOT_TOKEN:
    HEADERS["Authorization"] = PINOT_TOKEN

conn = connect(
    host=PINOT_BROKER_HOST,
    port=PINOT_BROKER_PORT,
    path="/query/sql",
    scheme=PINOT_BROKER_SCHEME,
    username=PINOT_USERNAME,
    password=PINOT_PASSWORD,
    use_multistage_engine=PINOT_USE_MSQE,
)


class Pinot:
    def __init__(self):
        self.insights: list[str] = []

    def _execute_query(self, query: str, params: dict[str, Any] | None = None) -> list[dict[str, Any]]:
        logger.debug(f"Executing query: {query}")
        curs = conn.cursor()
        curs.execute(query)
        df = pd.DataFrame(curs, columns=[item[0] for item in curs.description])
        return df.to_dict(orient="records")

    def _get_tables(self, params: dict[str, Any] | None = None) -> list[str]:
        url = f"{PINOT_CONTROLLER_URL}/tables"
        return requests.get(url, headers=HEADERS).json()["tables"]

    def _get_table_detail(self, tableName: str, params: dict[str, Any] | None = None) -> dict[str, Any]:
        url = f"{PINOT_CONTROLLER_URL}/tables/{tableName}/size"
        return requests.get(url, headers=HEADERS).json()

    def _get_segment__metadata_detail(self, tableName: str, params: dict[str, Any] | None = None) -> dict[str, Any]:
        url = f"{PINOT_CONTROLLER_URL}/segments/{tableName}/metadata"
        return requests.get(url, headers=HEADERS).json()

    def _get_segments(self, tableName: str, params: dict[str, Any] | None = None) -> dict[str, Any]:
        url = f"{PINOT_CONTROLLER_URL}/segments/{tableName}"
        return requests.get(url, headers=HEADERS).json()

    def _get_index_column_detail(self, tableName: str, segmentName: str, params: dict[str, Any] | None = None) -> dict[str, Any]:
        for type_suffix in ["REALTIME", "OFFLINE"]:
            url = f"{PINOT_CONTROLLER_URL}/segments/{tableName}_{type_suffix}/{segmentName}/metadata?columns=*"
            response = requests.get(url, headers=HEADERS)
            if response.status_code == 200:
                return response.json()
        raise ValueError("Index column detail not found")

    def _get_tableconfig_schema_detail(self, tableName: str, params: dict[str, Any] | None = None) -> dict[str, Any]:
        url = f"{PINOT_CONTROLLER_URL}/tableConfigs/{tableName}"
        return requests.get(url, headers=HEADERS).json()

    def list_tools(self) -> list[types.Tool]:
        return [
            types.Tool(
                name="read-query",
                description="Execute a SELECT query on the Pinot database",
                inputSchema={
                    "type": "object",
                    "properties": {
                        "query": {"type": "string", "description": "SELECT SQL query to execute"},
                    },
                    "required": ["query"],
                },
            ),
            types.Tool(name="list-tables", description="List all Pinot tables", inputSchema={"type": "object", "properties": {}}),
            types.Tool(
                name="table-details",
                description="Get details about a single table",
                inputSchema={
                    "type": "object",
                    "properties": {
                        "tableName": {"type": "string"},
                    },
                    "required": ["tableName"],
                },
            ),
            types.Tool(
                name="segment-list",
                description="List segments for a table",
                inputSchema={
                    "type": "object",
                    "properties": {
                        "tableName": {"type": "string"},
                    },
                    "required": ["tableName"],
                },
            ),
            types.Tool(
                name="index-column-details",
                description="Get index/column details",
                inputSchema={
                    "type": "object",
                    "properties": {
                        "tableName": {"type": "string"},
                        "segmentName": {"type": "string"},
                    },
                    "required": ["tableName", "segmentName"],
                },
            ),
            types.Tool(
                name="segment-metadata-details",
                description="Get metadata for segments",
                inputSchema={
                    "type": "object",
                    "properties": {
                        "tableName": {"type": "string"},
                    },
                    "required": ["tableName"],
                },
            ),
            types.Tool(
                name="tableconfig-schema-details",
                description="Get table config/schema",
                inputSchema={
                    "type": "object",
                    "properties": {
                        "tableName": {"type": "string"},
                    },
                    "required": ["tableName"],
                },
            ),
        ]

    def handle_tool(self, name: str, arguments: dict[str, Any]) -> list[types.TextContent]:
        match name:
            case "read-query":
                return [types.TextContent(type="text", text=str(self._execute_query(arguments["query"])))]
            case "list-tables":
                return [types.TextContent(type="text", text=str(self._get_tables()))]
            case "table-details":
                return [types.TextContent(type="text", text=str(self._get_table_detail(arguments["tableName"])))]
            case "segment-list":
                return [types.TextContent(type="text", text=str(self._get_segments(arguments["tableName"])))]
            case "index-column-details":
                return [types.TextContent(type="text", text=str(self._get_index_column_detail(arguments["tableName"], arguments["segmentName"])))]
            case "segment-metadata-details":
                return [types.TextContent(type="text", text=str(self._get_segment__metadata_detail(arguments["tableName"])))]
            case "tableconfig-schema-details":
                return [types.TextContent(type="text", text=str(self._get_tableconfig_schema_detail(arguments["tableName"])))]
            case _:
                raise ValueError(f"Unknown tool: {name}")
